import pika
import pika.exceptions
from kfsd.apps.core.common.logger import Logger, LogLevel
from kfsd.apps.core.common.kubefacets_config import KubefacetsConfig
from kfsd.apps.core.utils.http.django.config import DjangoConfig
from kfsd.apps.core.utils.dict import DictUtils
from kfsd.apps.core.common.singleton import Singleton

logger = Logger.getSingleton(__name__, LogLevel.DEBUG)


class RabbitMQ:
    PROPERTIES = "properties"
    BODY = "body"
    ON_MESSAGE_CALLBACK = "on_message_callback"

    def __init__(self, configId):
        self.__config = DjangoConfig(KubefacetsConfig().getConfig())
        self.__configId = configId
        self.prerun()

    @classmethod
    @Singleton
    def getSingleton(cls, configId):
        return cls(configId)

    def isConnectionOpen(self):
        return self.__connection.is_open

    def prerun(self):
        self.establishConnection()
        self.createChannel()
        self.declareExchanges()

    def establishConnection(self):
        self.__connection = self.connect()

    def createChannel(self):
        self.__channel = self.__connection.channel()

    def closeChannel(self):
        self.__channel.close()

    def getConfig(self):
        return self.__config

    def reconnect(self):
        logger.info("Reconnecting RabbitMQ..")
        self.establishConnection()
        self.createChannel()

    def connect(self):
        connectionConfig = self.__config.findConfigs(["services.rabbitmq.connect"])[
            0
        ].copy()
        logger.info("Connecting to RabbitMQ: {}".format(connectionConfig))
        authCredentials = connectionConfig.pop("credentials")
        connectionConfig["credentials"] = self.constructCredentials(
            DictUtils.get(authCredentials, "username"),
            DictUtils.get(authCredentials, "pwd"),
        )
        connection_params = pika.ConnectionParameters(**connectionConfig)
        return pika.BlockingConnection(connection_params)

    def constructCredentials(self, username, pwd):
        return pika.PlainCredentials(username, pwd)

    def declareExchanges(self):
        exchanges = self.__config.findConfigs(["services.rabbitmq.exchanges"])[0]
        for exchangeAttrs in exchanges:
            self.__channel.exchange_declare(**exchangeAttrs)

    def declareQueues(self):
        declarations = self.__config.findConfigs(
            ["services.rabbitmq.consumers.declarations"]
        )[0]
        for queueAttrs in declarations:
            queueAttrs["queue"] = queueAttrs["queue"].format(service_id=self.__configId)
            self.__channel.queue_declare(**queueAttrs)

    def bindQueues(self):
        binds = self.__config.findConfigs(["services.rabbitmq.consumers.binds"])[0]
        for bindAttrs in binds:
            if "identifier" in bindAttrs:
                bindAttrs.pop("identifier")

            bindAttrs["queue"] = bindAttrs["queue"].format(service_id=self.__configId)
            bindAttrs["routing_key"] = bindAttrs["routing_key"].format(
                service_id=self.__configId
            )
            self.__channel.queue_bind(**bindAttrs)

    def consumeQueues(self, callback):
        self.declareQueues()
        self.bindQueues()
        queues = self.__config.findConfigs(["services.rabbitmq.consumers.queues"])[0]
        for queueAttrs in queues:
            queueAttrs["queue"] = queueAttrs["queue"].format(service_id=self.__configId)
            self.consume(queueAttrs, callback)

    def publish(self, attrs, msg):
        if not self.isConnectionOpen():
            self.reconnect()
        attrs = DictUtils.deepcopy(attrs)
        attrs[self.BODY] = msg
        properties = DictUtils.get(attrs, self.PROPERTIES, None)
        if properties:
            properties = pika.BasicProperties(**properties)
            attrs[self.PROPERTIES] = properties

        self.__channel.basic_publish(**attrs)

    def consume(self, attrs, callback):
        attrs[self.ON_MESSAGE_CALLBACK] = callback
        self.__channel.basic_consume(**attrs)

    def closeConnection(self):
        logger.info("Closing RabbitMQ connection!")
        self.__connection.close()

    def startConsuming(self):
        self.__channel.start_consuming()


def close_rabbitmq_connection(self):
    try:
        self.closeConnection()
    except Exception as e:
        logger.error("Could not close RabbitMQ connection!: {}".format(e))



from django.core.management.base import BaseCommand
from kfsd.apps.core.msmq.rabbitmq.base import RabbitMQ
from kfsd.apps.core.utils.time import Time
from kfsd.apps.endpoints.serializers.common.outpost import MsgSerializer

from kfsd.apps.core.common.logger import Logger, LogLevel
from kfsd.apps.core.utils.dict import DictUtils
from kfsd.apps.models.tables.outpost import Outpost, send_msg
from kfsd.apps.models.constants import (
    OUTPOST_ACTION_CLEAR,
    USER_ACTION_CREATE,
    USER_ACTION_DELETE,
    USER_ACTION_UPDATE,
)

from kfsd.apps.models.tables.reference import Reference
from kfsd.apps.endpoints.serializers.common.reference import (
    ReferenceViewModelSerializer,
)


import json

logger = Logger.getSingleton(__name__, LogLevel.DEBUG)


def clear_outpost(data):
    outpostQS = Outpost.objects.filter(status="E").order_by("created")
    identifiers = [outpost.identifier for outpost in outpostQS]
    logger.info(
        "Recd CLEAR_OUTPOST command on msmq, identifiers: {} with 'Error' status found.".format(
            identifiers if identifiers else None
        )
    )
    for outpostIns in outpostQS:
        send_msg(outpostIns)


def construct_ref(sourceModel, data):
    identifier = data.pop("identifier")
    return {"identifier": identifier, "type": sourceModel, "attrs": data}


def create_ref(source, data):
    idSerailizer = ReferenceViewModelSerializer(data=construct_ref(source, data))
    idSerailizer.is_valid(raise_exception=True)
    idSerailizer.save()


def update_ref(source, data):
    refData = construct_ref(source, data)
    identifier = DictUtils.get(refData, "identifier")
    instance = Reference.objects.get(identifier=identifier)
    serializedData = ReferenceViewModelSerializer(instance, data=refData, partial=True)
    serializedData.is_valid(raise_exception=True)
    serializedData.save()


def delete_ref(source, data):
    refData = construct_ref(source, data)
    identifier = DictUtils.get(refData, "identifier")
    instance = Reference.objects.get(identifier=identifier)
    instance.delete()


def action_map_to_source_model(action):
    if action in [USER_ACTION_CREATE, USER_ACTION_UPDATE, USER_ACTION_DELETE]:
        return "USER"


def process_ref(requestData):
    action = DictUtils.get_by_path(requestData, "action")
    data = DictUtils.get_by_path(requestData, "data")
    targetModel = DictUtils.get(requestData, "target_model", "")
    logger.info(
        "Recd {} command on msmq, request data: {}".format(
            action, json.dumps(requestData, indent=4)
        )
    )
    if action in [
        USER_ACTION_CREATE,
    ]:
        create_ref(targetModel, data)

    if action in [
        USER_ACTION_UPDATE,
    ]:
        update_ref(targetModel, data)

    if action in [
        USER_ACTION_DELETE,
    ]:
        delete_ref(targetModel, data)


callback_map = {
    OUTPOST_ACTION_CLEAR: clear_outpost,
    USER_ACTION_CREATE: process_ref,
    USER_ACTION_UPDATE: process_ref,
    USER_ACTION_DELETE: process_ref,
}


def base_callback(ch, method, properties, body):
    bodyStr = body.decode().replace("'", '"')
    jsonStr = json.loads(bodyStr)
    serializedData = MsgSerializer(data=jsonStr)
    serializedData.is_valid(raise_exception=True)

    action = DictUtils.get(serializedData.data, "action")
    if action in callback_map:
        callback_map[action](serializedData.data)
    else:
        logger.error("Action : {} not handled in message consumption".format(action))


class Command(BaseCommand):
    help = "Listens to a RabbitMQ topic"

    def __init__(self, callbackFn=base_callback):
        self.__callbackFn = callbackFn

    def add_arguments(self, parser):
        parser.add_argument(
            "-s",
            "--service_config_id",
            type=str,
            help="Service Config Id",
        )

    def connectToMSMQ(self, serviceConfigId):
        try:
            msmqHandler = RabbitMQ.getSingleton(serviceConfigId)
            return msmqHandler
        except Exception as e:
            print(e)
            logger.error(
                "Error connecting to RabbitMQ, check if RabbitMQ instance is up!"
            )
            Time.sleep(30)
            self.connectToMSMQ()

    def handle(self, *args, **options):
        logger.info("Listening to MSMQ messages...")
        serviceConfigId = DictUtils.get(options, "service_config_id")
        msmqHandler = self.connectToMSMQ(serviceConfigId)
        msmqHandler.consumeQueues(self.__callbackFn)
        msmqHandler.startConsuming()



from django.db import models
from django.utils.text import slugify

from kfsd.apps.models.constants import MAX_LENGTH
from kfsd.apps.models.tables.base import BaseModel


# Github, Twitter, Linkedin, Youtube, Website
class Platform(BaseModel):
    name = models.CharField(max_length=MAX_LENGTH)
    type = models.CharField(max_length=MAX_LENGTH)
    slug = models.SlugField()

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name)
        self.identifier = "TYPE={},PLATFORM={}".format(self.type, self.name)
        return super().save(*args, **kwargs)

    class Meta:
        app_label = "backend"
        verbose_name = "Platform"
        verbose_name_plural = "Platform"



from drf_spectacular.utils import extend_schema_view

from kfsd.apps.endpoints.views.common.custom_model import CustomModelViewSet
from kfsd.apps.models.tables.platform import Platform
from kfsd.apps.endpoints.serializers.common.platform import PlatformViewModelSerializer
from kfsd.apps.endpoints.views.docs.platform import PlatformDoc


@extend_schema_view(**PlatformDoc.modelviewset())
class PlatformModelViewSet(CustomModelViewSet):
    queryset = Platform.objects.all()
    serializer_class = PlatformViewModelSerializer



from drf_spectacular.utils import extend_schema

from kfsd.apps.endpoints.views.docs.v1.platform import PlatformV1Doc
from kfsd.apps.endpoints.serializers.common.platform import PlatformViewModelSerializer
from kfsd.apps.endpoints.serializers.base import (
    NotFoundSerializer,
    ErrorSerializer,
)


class PlatformDoc:
    @staticmethod
    def modelviewset():
        return {
            "list": extend_schema(**PlatformDoc.modelviewset_list()),
            "retrieve": extend_schema(**PlatformDoc.modelviewset_get()),
            "destroy": extend_schema(**PlatformDoc.modelviewset_delete()),
            "partial_update": extend_schema(**PlatformDoc.modelviewset_patch()),
            "create": extend_schema(**PlatformDoc.modelviewset_create()),
        }

    @staticmethod
    def modelviewset_patch():
        return {
            "summary": "Platform - Patch",
            "description": "Platform Patch",
            "tags": ["PLATFORM"],
            "responses": {
                200: PlatformViewModelSerializer,
                404: NotFoundSerializer,
                500: ErrorSerializer,
            },
            "parameters": PlatformV1Doc.modelviewset_patch_path_examples(),
            "examples": PlatformV1Doc.modelviewset_patch_examples(),
        }

    @staticmethod
    def modelviewset_list():
        return {
            "summary": "Platform - List",
            "description": "Platform - All",
            "tags": ["PLATFORM"],
            "responses": {
                200: PlatformViewModelSerializer,
                404: NotFoundSerializer,
                500: ErrorSerializer,
            },
            "parameters": PlatformV1Doc.modelviewset_list_path_examples(),
            "examples": PlatformV1Doc.modelviewset_list_examples(),
        }

    @staticmethod
    def modelviewset_get():
        return {
            "summary": "Platform - Get",
            "description": "Platform Detail",
            "tags": ["PLATFORM"],
            "responses": {
                200: PlatformViewModelSerializer,
                404: NotFoundSerializer,
                500: ErrorSerializer,
            },
            "parameters": PlatformV1Doc.modelviewset_get_path_examples(),
            "examples": PlatformV1Doc.modelviewset_get_examples(),
        }

    @staticmethod
    def modelviewset_create():
        return {
            "summary": "Platform - Create",
            "description": "Platform - Create",
            "tags": ["PLATFORM"],
            "responses": {
                200: PlatformViewModelSerializer,
                400: ErrorSerializer,
                404: ErrorSerializer,
                500: ErrorSerializer,
            },
            "examples": PlatformV1Doc.modelviewset_create_examples(),
        }

    @staticmethod
    def modelviewset_delete():
        return {
            "summary": "Platform - Delete",
            "description": "Platform Delete",
            "tags": ["PLATFORM"],
            "responses": {204: None, 404: NotFoundSerializer, 500: ErrorSerializer},
            "parameters": PlatformV1Doc.modelviewset_delete_path_examples(),
        }



from rest_framework import serializers
from django.core.validators import (
    MinLengthValidator,
)

from kfsd.apps.models.constants import MIN_LENGTH
from kfsd.apps.endpoints.serializers.common.model import BaseModelSerializer
from kfsd.apps.models.tables.link import Link
from kfsd.apps.models.tables.platform import Platform


class LinkModelSerializer(BaseModelSerializer):
    link = serializers.CharField(
        validators=[
            MinLengthValidator(MIN_LENGTH),
        ]
    )
    platform = serializers.SlugRelatedField(
        required=False,
        many=False,
        read_only=False,
        slug_field="identifier",
        queryset=Platform.objects.all(),
    )
    link_id = serializers.CharField(read_only=True)

    class Meta:
        model = Link
        fields = "__all__"


class LinkViewModelSerializer(BaseModelSerializer):
    id = None
    created = None
    updated = None

    class Meta:
        model = Link
        exclude = ("created", "updated", "id")



from rest_framework import serializers
from django.core.validators import (
    MinLengthValidator,
    MaxLengthValidator,
)

from kfsd.apps.models.constants import MAX_LENGTH, MIN_LENGTH
from kfsd.apps.endpoints.serializers.common.model import BaseModelSerializer
from kfsd.apps.models.tables.platform import Platform


class PlatformModelSerializer(BaseModelSerializer):
    name = serializers.CharField(
        validators=[
            MinLengthValidator(MIN_LENGTH),
            MaxLengthValidator(MAX_LENGTH),
        ]
    )
    type = serializers.CharField(
        validators=[
            MinLengthValidator(MIN_LENGTH),
            MaxLengthValidator(MAX_LENGTH),
        ]
    )
    slug = serializers.SlugField(required=False)

    class Meta:
        model = Platform
        fields = "__all__"


class PlatformViewModelSerializer(BaseModelSerializer):
    id = None
    created = None
    updated = None

    class Meta:
        model = Platform
        exclude = ("created", "updated", "id")



from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
import json
import pika.exceptions

from kfsd.apps.models.tables.base import BaseModel
from kfsd.apps.core.utils.system import System
from kfsd.apps.core.utils.time import Time
from kfsd.apps.core.msmq.rabbitmq.base import RabbitMQ
from kfsd.apps.core.common.logger import Logger, LogLevel


logger = Logger.getSingleton(__name__, LogLevel.DEBUG)


class Outpost(BaseModel):
    STATUS_CHOICES = (
        ("P", "PENDING"),
        ("I", "IN-PROGRESS"),
        ("E", "ERROR"),
        ("C", "COMPLETED"),
    )

    msg_queue_info = models.JSONField(default=dict)
    msg = models.JSONField(default=dict)
    status = models.CharField(max_length=1, choices=STATUS_CHOICES, default="P")
    attempts = models.IntegerField(default=0)
    debug_info = models.JSONField(default=dict)

    def save(self, *args, **kwargs):
        if not self.identifier:
            self.identifier = System.uuid(32)
        return super().save(*args, **kwargs)

    class Meta:
        app_label = "models"
        verbose_name = "Outpost"
        verbose_name_plural = "Outpost"


@receiver(post_save, sender=Outpost)
def signal_send_msg(sender, instance, created, **kwargs):
    if created:
        send_msg(instance)


def log_error(instance):
    oldDebugInfo = instance.debug_info
    currentTimeStr = Time.calculate_time(Time.current_time(), {"minutes": 0}, True)
    oldDebugInfo[currentTimeStr] = "Error occurred in publishing outpost msg"
    instance.debug_info = oldDebugInfo
    instance.attempts += 1
    instance.status = "E"
    instance.save()


def send_msg(instance):
    logger.info(
        "Sending msg to MSMQ for outpost id: {}, msmq: {}, msg: {}".format(
            instance.identifier,
            json.dumps(instance.msg_queue_info, indent=4),
            json.dumps(instance.msg, indent=4),
        )
    )
    try:
        configId = System.getEnv("service_id")
        rabbitMQ = RabbitMQ.getSingleton(configId)
        rabbitMQ.publish(instance.msg_queue_info, json.dumps(instance.msg))
        instance.delete()
    except pika.exceptions.StreamLostError as e:
        logger.error("Recd error on publish: {}".format(e.__str__()))
        log_error(instance)
    except Exception as e:
        logger.error("Recd error on publish: {}".format(e.__str__()))
        log_error(instance)



from drf_spectacular.utils import extend_schema_view

from kfsd.apps.endpoints.views.common.custom_model import CustomModelViewSet
from kfsd.apps.models.tables.reference import Reference
from kfsd.apps.endpoints.serializers.common.reference import (
    ReferenceViewModelSerializer,
)
from kfsd.apps.endpoints.views.docs.reference import ReferenceDoc


@extend_schema_view(**ReferenceDoc.modelviewset())
class ReferenceModelViewSet(CustomModelViewSet):
    queryset = Reference.objects.all()
    serializer_class = ReferenceViewModelSerializer
